"""Custom Morphology classes."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unnecessary-comprehension
import logging

from bluepyopt import ephys

logger = logging.getLogger(__name__)


class SSCXNrnFileMorphology(ephys.morphologies.NrnFileMorphology):
    """Custom Morphology.

    Attributes:
        name (str): name of this object
        comment (str): comment
        morphology_path (str): location of the file describing the morphology
        do_replace_axon (bool): Does the axon need to be replaced by an AIS
            stub with default function ?
        replace_axon_hoc (str): Translation in HOC language for the
            'replace_axon' method. This code will 'only' be used when
            calling create_hoc on a cell model. While the model is run in
            python, replace_axon is used instead. Must include
            'proc replace_axon(){ ... }
            If None,the default replace_axon is used
        do_set_nseg (bool): if True, it will use nseg_frequency
        nseg_frequency (float): frequency of nseg
        morph_modifiers (list): list of functions to modify the icell
            with (sim, icell) as arguments
        morph_modifiers_hoc (list): list of hoc strings corresponding
            to morph_modifiers
    """

    def replace_axon(self, sim=None, icell=None):
        """Replace axon.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        L_target = 60  # length of stub axon
        nseg0 = 5  # number of segments for each of the two axon sections

        nseg_total = nseg0 * 2
        chunkSize = L_target / nseg_total

        diams = []
        lens = []

        count = 0
        for section in icell.axonal:
            L = section.L
            nseg = 1 + int(L / chunkSize / 2.0) * 2  # nseg to get diameter
            section.nseg = nseg

            for seg in section:
                count = count + 1
                diams.append(seg.diam)
                lens.append(L / nseg)
                if count == nseg_total:
                    break
            if count == nseg_total:
                break

        for section in icell.axonal:
            sim.neuron.h.delete_section(sec=section)

        #  new axon array
        sim.neuron.h.execute("create axon[2]", icell)

        L_real = 0
        count = 0

        for section in icell.axon:
            section.nseg = int(nseg_total / 2)
            section.L = L_target / 2

            for seg in section:
                seg.diam = diams[count]
                L_real = L_real + lens[count]
                count = count + 1

            icell.axonal.append(sec=section)
            icell.all.append(sec=section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

        sim.neuron.h.execute("create myelin[1]", icell)
        icell.myelinated.append(sec=icell.myelin[0])
        icell.all.append(sec=icell.myelin[0])
        icell.myelin[0].nseg = 5
        icell.myelin[0].L = 1000
        icell.myelin[0].diam = diams[count - 1]
        icell.myelin[0].connect(icell.axon[1], 1.0, 0.0)

        logger.debug(
            "Replace axon with tapered AIS of length %d, target length was %d, diameters are %s",
            L_real,
            L_target,
            diams,
        )


class ThalamusNrnFileMorphology(ephys.morphologies.NrnFileMorphology):
    """Custom Morphology.

    Attributes:
        name (str): name of this object
        comment (str): comment
        morphology_path (str): location of the file describing the morphology
        do_replace_axon (bool): Does the axon need to be replaced by an AIS
            stub with default function ?
        replace_axon_hoc (str): Translation in HOC language for the
            'replace_axon' method. This code will 'only' be used when
            calling create_hoc on a cell model. While the model is run in
            python, replace_axon is used instead. Must include
            'proc replace_axon(){ ... }
            If None,the default replace_axon is used
        do_set_nseg (bool): if True, it will use nseg_frequency
        nseg_frequency (float): frequency of nseg
        morph_modifiers (list): list of functions to modify the icell
            with (sim, icell) as arguments
        morph_modifiers_hoc (list): list of hoc strings corresponding
            to morph_modifiers
    """

    @staticmethod
    def replace_axon(sim=None, icell=None):
        """Replace axon.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        L_target = 60  # length of stub axon
        nseg0 = 5  # number of segments for each of the two axon sections

        nseg_total = nseg0 * 2
        chunkSize = L_target / nseg_total

        diams = []
        lens = []

        count = 0
        for section in icell.axonal:
            L = section.L
            nseg = 1 + int(L / chunkSize / 2.0) * 2  # nseg to get diameter
            section.nseg = nseg

            for seg in section:
                diams.append(seg.diam)
                lens.append(L / nseg)
                if count == nseg_total:
                    break
            if count == nseg_total:
                break

        # Work-around if axon is too short
        lasti = -2  # Last diam. may be bigger if axon cut

        if len(diams) < nseg_total:
            diams = diams + [diams[lasti]] * (nseg_total - len(diams))
            lens = lens + [lens[lasti]] * (nseg_total - len(lens))
            if nseg_total - len(diams) > 5:
                logger.debug(
                    "Axon too short, adding more than 5 sections with fixed diam"
                )

        for section in icell.axonal:
            sim.neuron.h.delete_section(sec=section)

        #  new axon array
        sim.neuron.h.execute("create axon[2]", icell)

        L_real = 0
        count = 0

        for section in icell.axon:
            section.nseg = int(nseg_total / 2)
            section.L = L_target / 2

            for seg in section:
                seg.diam = diams[count]
                L_real = L_real + lens[count]
                count = count + 1

            icell.axonal.append(sec=section)
            icell.all.append(sec=section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

        logger.debug(
            "Replace axon with tapered AIS of length %d, target length was %d, diameters are %s",
            L_real,
            L_target,
            diams,
        )



"""Json utilities."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import numpy as np


class NpEncoder(json.JSONEncoder):
    """Class to encode numpy object as python object."""

    def default(self, o):
        """Convert numpy values to regular python values.

        Args:
            o: value to enventually convert if is a numpy object

        Returns:
            converted values or parent default method if not a numpy object
        """
        if isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        else:
            return super(NpEncoder, self).default(o)



"""Protocol creation functions & custom protocol classes."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

from emodelrunner.create_recordings import get_pairsim_recordings
from emodelrunner.create_stimuli import load_pulses
from emodelrunner.configuration import PackageType
from emodelrunner.protocols import synplas_protocols

from emodelrunner.synapses.recordings import SynapseRecordingCustom
from emodelrunner.stimuli import MultipleSteps
from emodelrunner.features import define_efeatures
from emodelrunner.protocols.reader import ProtocolParser
from emodelrunner.synapses.create_locations import get_syn_locs
from emodelrunner.synapses.stimuli import (
    NrnVecStimStimulusCustom,
    NetConSpikeDetector,
)

logger = logging.getLogger(__name__)


class ProtocolBuilder:
    """Class representing the protocols applied in SSCX.

    Attributes:
        protocols (bluepyopt.ephys.protocols.SequenceProtocol): the protocols to apply to the cell
    """

    def __init__(self, protocols):
        """Constructor to be called by the classmethod overloads.

        Args:
            protocols (bluepyopt.ephys.protocols.SequenceProtocol): protocols to apply to the cell
        """
        self.protocols = protocols

    @staticmethod
    def _get_syn_locs(add_synapses, cell):
        """Wraps get_syn_locs with exception raising."""
        syn_locs = None
        if add_synapses:
            if cell is not None:
                syn_locs = get_syn_locs(cell)
            else:
                raise RuntimeError("'None' value encountered in cell object.")
        return syn_locs

    @classmethod
    def using_sscx_protocols(cls, add_synapses, prot_args, cell=None):
        """Creates the object with the sscx protocols.

        Args:
            add_synapses (bool): whether to add synapses to the cell
            prot_args (dict): config data relative to protocols
                See load.get_prot_args for details
            cell (CellModelCustom): cell model
        Returns:
            ProtocolBuilder: the object with the sscx protocols
        """
        syn_locs = cls._get_syn_locs(add_synapses, cell)

        protocols = create_protocols_object(
            apical_point_isec=prot_args["apical_point_isec"],
            prot_path=prot_args["prot_path"],
            package_type=PackageType.sscx,
            features_path=prot_args["features_path"],
            mtype=prot_args["mtype"],
            syn_locs=syn_locs,
        )
        return cls(protocols)

    @classmethod
    def using_thalamus_protocols(cls, add_synapses, prot_args, cell=None):
        """Creates the object with the thalamus protocols.

        Args:
            add_synapses (bool): whether to add synapses to the cell
            prot_args (dict): config data relative to protocols
                See load.get_prot_args for details
            cell (CellModelCustom): cell model
        Returns:
            ProtocolBuilder: the object with the thalamus protocols
        """
        syn_locs = cls._get_syn_locs(add_synapses, cell)

        protocols = create_protocols_object(
            apical_point_isec=prot_args["apical_point_isec"],
            prot_path=prot_args["prot_path"],
            package_type=PackageType.thalamus,
            features_path=prot_args["features_path"],
            mtype=prot_args["mtype"],
            syn_locs=syn_locs,
        )
        return cls(protocols)

    def get_ephys_protocols(self):
        """Returns the list of ephys protocol objects.

        Returns:
            bluepyopt.ephys.protocols.SequenceProtocol: the protocols to apply to the cell
        """
        return self.protocols

    def get_stim_currents(self, responses, dt):
        """Generates the currents injected by protocols.

        Args:
            responses (dict): the responses to the protocols run
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict: the currents of the protocols.

            If the MainProtocol was used, only the RMP protocol,
            'pre-protocols' and 'other protocols' currents are returned
        """
        # find threshold and holding currents
        thres_i = None
        hold_i = None
        for key, resp in responses.items():
            if "threshold_current" in key:
                thres_i = resp
            elif "holding_current" in key:
                hold_i = resp

        currents = {}
        for protocol in self.protocols.protocols:
            currents.update(
                protocol.generate_current(
                    threshold_current=thres_i, holding_current=hold_i, dt=dt
                )
            )

        return currents

    def get_thalamus_stim_currents(self, responses, mtype, dt):
        """Returns the currents injected by thalamus protocols.

        Args:
            responses (dict): protocol responses
            mtype (str): mtype to index the responses
            dt (float): timestep of the generated currents (ms)

        Returns:
            dict: the generated currents in a dict.
        """
        thres_i_hyp = responses[f"{mtype}.bpo_threshold_current_hyp"]
        thres_i_dep = responses[f"{mtype}.bpo_threshold_current_dep"]
        holding_i_hyp = responses[f"{mtype}.bpo_holding_current_hyp"]
        holding_i_dep = responses[f"{mtype}.bpo_holding_current_dep"]
        currents = {}
        for protocol in self.protocols.protocols:
            currents.update(
                protocol.generate_current(
                    thres_i_hyp, thres_i_dep, holding_i_hyp, holding_i_dep, dt
                )
            )
        return currents


def create_protocols_object(
    apical_point_isec,
    prot_path,
    package_type,
    features_path="",
    mtype="",
    syn_locs=None,
    stochkv_det=None,
):
    """Return a dict containing protocols.

    Args:
        apical_point_isec (int): section index of the apical point
            Set to -1 no apical point is used in any extra recordings
        prot_path (str): path to the protocols file
        package_type (Enum): enum denoting the package type
        features_path (str): path to the features file
        mtype (str): morphology name to be used as prefix in output filenames
        syn_locs (list): list of synapse locations
        stochkv_det (bool): set if stochastic or deterministic

    Raises:
        ValueError: if the package type is not supported

    Returns:
        ephys.protocols.SequenceProtocol: sequence protocol containing all the protocols
    """
    # pylint: disable=unbalanced-tuple-unpacking, too-many-locals
    if package_type == PackageType.sscx:
        protocols_dict = ProtocolParser().parse_sscx_protocols(
            prot_path,
            stochkv_det,
            mtype,
            apical_point_isec,
            syn_locs,
        )
    elif package_type == PackageType.thalamus:
        protocols_dict = ProtocolParser().parse_thalamus_protocols(
            prot_path,
            stochkv_det,
            mtype,
        )
    else:
        raise ValueError(f"unsupported package type: {package_type}")

    if "Main" in protocols_dict:
        efeatures = define_efeatures(
            protocols_dict["Main"],
            features_path,
            mtype,
        )

        if package_type == PackageType.sscx:
            set_sscx_main_protocol_efeatures(protocols_dict, efeatures, prefix=mtype)
        elif package_type == PackageType.thalamus:
            set_thalamus_main_protocol_efeatures(
                protocols_dict, efeatures, prefix=mtype
            )

        protocols = [protocols_dict["Main"]]
    else:
        protocols = list(protocols_dict.values())

    return ephys.protocols.SequenceProtocol(
        "all protocols",
        protocols=protocols,
    )


def set_sscx_main_protocol_efeatures(protocols_dict, efeatures, prefix):
    """Set the efeatures of the main protocol.

    Args:
        protocols_dict (dict): contains all protocols to be run
            If this function is called, should contain the MainProtocol
            and the associated protocols (RinHoldCurrent, ThresholdDetection)
        efeatures (dict): contains the efeatures
        prefix (str): prefix used in naming responses, features, recordings, etc.
    """
    protocols_dict["Main"].rmp_efeature = efeatures[f"{prefix}.RMP.soma.v.voltage_base"]

    protocols_dict["Main"].rin_efeature = efeatures[
        f"{prefix}.Rin.soma.v.ohmic_input_resistance_vb_ssse"
    ]

    protocols_dict["Main"].rin_efeature.stimulus_current = protocols_dict[
        "Main"
    ].rinhold_protocol.rin_protocol_template.step_amplitude

    protocols_dict["RinHoldcurrent"].voltagebase_efeature = efeatures[
        f"{prefix}.Rin.soma.v.voltage_base"
    ]
    protocols_dict["ThresholdDetection"].holding_voltage = efeatures[
        f"{prefix}.Rin.soma.v.voltage_base"
    ].exp_mean


def set_thalamus_main_protocol_efeatures(protocols_dict, efeatures, prefix):
    """Set the efeatures of the main thalamus protocol.

    Args:
        protocols_dict (dict): contains all protocols to be run
            If this function is called, should contain the MainProtocol
            and the associated protocols (RinHoldCurrent, ThresholdDetection)
        efeatures (dict): contains the efeatures
        prefix (str): prefix used in naming responses, features, recordings, etc.
    """
    protocols_dict["Main"].rmp_efeature = efeatures[
        f"{prefix}.RMP.soma.v.steady_state_voltage_stimend"
    ]

    try:
        protocols_dict["Main"].rin_efeature_dep = efeatures[
            f"{prefix}.Rin_dep.soma.v.ohmic_input_resistance_vb_ssse"
        ]
    except KeyError:
        pass
    try:
        protocols_dict["Main"].rin_efeature_dep.stimulus_current = protocols_dict[
            "Main"
        ].rinhold_protocol_dep.rin_protocol_template.step_stimulus.step_amplitude
    except AttributeError:
        pass

    protocols_dict["Main"].rin_efeature_hyp = efeatures[
        f"{prefix}.Rin_hyp.soma.v.ohmic_input_resistance_vb_ssse"
    ]

    protocols_dict["Main"].rin_efeature_hyp.stimulus_current = protocols_dict[
        "Main"
    ].rinhold_protocol_hyp.rin_protocol_template.step_stimulus.step_amplitude

    try:
        protocols_dict["RinHoldcurrent_dep"].voltagebase_efeature = efeatures[
            f"{prefix}.Rin_dep.soma.v.voltage_base"
        ]
        protocols_dict["ThresholdDetection_dep"].holding_voltage = efeatures[
            f"{prefix}.Rin_dep.soma.v.voltage_base"
        ].exp_mean

    except KeyError:
        pass
    protocols_dict["RinHoldcurrent_hyp"].voltagebase_efeature = efeatures[
        f"{prefix}.Rin_hyp.soma.v.voltage_base"
    ]
    protocols_dict["ThresholdDetection_hyp"].holding_voltage = efeatures[
        f"{prefix}.Rin_hyp.soma.v.voltage_base"
    ].exp_mean


def define_synapse_plasticity_protocols(
    cell,
    pre_spike_train,
    protocol_name,
    cvode_active,
    synrecs,
    tstop,
    fastforward,
    stim_path="protocols/stimuli.json",
):
    """Create stimuli and protocols to run glusynapse cell.

    Args:
        cell (CellModelCustom): post-synaptic cell model
        pre_spike_train (list): times at which the synapses fire (ms)
        protocol_name (str): protocol name to be passed on to Recording and Protocol classes
        cvode_active (bool): whether to activate the adaptative time step
        synrecs (list of str): the extra synapse variables to record
        tstop (float): total duration of the simulation (ms)
        fastforward (float): time at which to enable synapse fast-forwarding (ms)
        stim_path (str): path to the pulse stimuli file

    Returns:
        synplas_protocols.SweepProtocolCustom: synapse plasticity protocols
    """
    # pylint: disable=too-many-locals
    syn_locs = get_syn_locs(cell)
    syn_stim = NrnVecStimStimulusCustom(
        syn_locs,
        stop=tstop,
        pre_spike_train=pre_spike_train,
    )

    # recording location
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5
    )
    # recording
    rec = ephys.recordings.CompRecording(
        name=protocol_name, location=soma_loc, variable="v"
    )
    recs = [rec]

    for syn_loc in syn_locs:
        for synrec in synrecs:
            recs.append(
                SynapseRecordingCustom(name=synrec, location=syn_loc, variable=synrec)
            )

    # pulses
    stims = load_pulses(soma_loc, stim_path)

    stims.append(syn_stim)

    # create protocol
    return synplas_protocols.SweepProtocolCustom(
        protocol_name, stims, recs, cvode_active, fastforward
    )


def define_pairsim_protocols(
    postcell,
    presyn_prot_name,
    postsyn_prot_name,
    cvode_active,
    synrecs,
    tstop,
    fastforward,
    presyn_stim_args,
    stim_path="protocols/stimuli.json",
):
    """Create stimuli and protocols to run glusynapse cell.

    Args:
        postcell (CellModelCustom): post-synaptic cell model
        presyn_prot_name (str): presynaptic protocol name
            to be passed on to Recording and Protocol classes
        postsyn_prot_name (str): postsynaptic protocol name
            to be passed on to Recording and Protocol classes
        cvode_active (bool): whether to activate the adaptative time step
        synrecs (list of str): the extra synapse variables to record
        tstop (float): total duration of the simulation (ms)
        fastforward (float): time at which to enable synapse fast-forwarding (ms)
        presyn_stim_args (dict): presynaptic stimulus configuration data
            See load.get_presyn_stim_args for details
        stim_path (str): path to the pulse stimuli file

    Returns:
        synplas_protocols.SweepProtocolPairSim: pair simulation synapse plasticity protocols
    """
    # pylint: disable=too-many-arguments, too-many-locals
    # locations
    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5
    )
    syn_locs = get_syn_locs(postcell)

    # recordings
    # has the structure (precell_recs, postcell_recs)
    recs = get_pairsim_recordings(
        soma_loc, syn_locs, synrecs, presyn_prot_name, postsyn_prot_name
    )

    # stimuli
    postsyn_stims = load_pulses(soma_loc, stim_path)

    presyn_stims = [
        MultipleSteps(
            soma_loc,
            presyn_stim_args["stim_train"],
            presyn_stim_args["amp"],
            presyn_stim_args["width"],
        )
    ]

    # appened to presyn stim because the precell is needed
    # to activate the postcell synapses
    syn_stim = NetConSpikeDetector(total_duration=tstop, locations=syn_locs)
    presyn_stims.append(syn_stim)

    # create protocol
    protocol_name = (
        f"presynaptic protocol: {presyn_prot_name},"
        + f"postsynaptic protocol: {postsyn_prot_name}"
    )
    return synplas_protocols.SweepProtocolPairSim(
        protocol_name,
        (presyn_stims, postsyn_stims),
        recs,
        cvode_active,
        fastforward,
    )



"""Synapse Point Process Mechanisms."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bluepyopt import ephys

from emodelrunner.synapses.glusynapse import GluSynapseCustom
from emodelrunner.synapses.synapse import SynapseCustom


class NrnMODPointProcessMechanismCustom(ephys.mechanisms.Mechanism):
    """Class containing all the synapses.

    Attributes:
        synapses_data (list of dicts): synapse data
        synconf_dict (dict): synapse configuration
        seed (int): random number generator seed number
        rng_settings_mode (str): mode of the random number generator
                Can be "Random123" or "Compatibility"
        pre_mtypes (list of ints): activate only synapses whose pre_mtype
            is in this list if None, all synapses are activated
        stim_params (dict or None): dict with pre_mtype as key,
            and netstim params list as item.
            netstim params list is [start, interval, number, noise]
        use_glu_synapse (bool): if True, instantiate synapses to use GluSynapse
        syn_setup_params (dict): contains extra parameters to setup synapses
            when using GluSynapseCustom
        rng (neuron Random): random number generator of the simulator
        pprocesses (list of SynapseCustom or GluSynapseCustom): list of the synapses
    """

    def __init__(
        self,
        name,
        synapses_data,
        synconf_dict,
        seed,
        rng_settings_mode,
        pre_mtypes=None,
        stim_params=None,
        comment="",
        use_glu_synapse=False,
        syn_setup_params=None,
    ):
        """Constructor.

        Args:
            name (str): name of this object
            synapses_data (list of dicts): synapse data
            synconf_dict (dict): synapse configuration
            seed (int): random number generator seed number
            rng_settings_mode (str): mode of the random number generator
                 Can be "Random123" or "Compatibility"
            pre_mtypes (list of ints): activate only synapses whose pre_mtype
                is in this list if None, all synapses are activated
            stim_params (dict or None): dict with pre_mtype as key,
                and netstim params list as item.
                netstim params list is [start, interval, number, noise]
            comment (str): comment
            use_glu_synapse (bool): if True, instantiate synapses to use GluSynapse
            syn_setup_params (dict): contains extra parameters to setup synapses
                when using GluSynapseCustom
        """
        # pylint: disable=too-many-arguments
        super(NrnMODPointProcessMechanismCustom, self).__init__(name, comment)
        self.synapses_data = synapses_data
        self.synconf_dict = synconf_dict
        self.seed = seed
        self.rng_settings_mode = rng_settings_mode
        self.pre_mtypes = pre_mtypes
        self.stim_params = stim_params
        self.use_glu_synapse = use_glu_synapse
        self.syn_setup_params = syn_setup_params
        self.rng = None
        self.pprocesses = None

    @staticmethod
    def get_cell_section_for_synapse(synapse, icell):
        """Returns the cell section on which is the synapse.

        Args:
            synapse (dict): contains the synapse data
            icell (neuron cell): cell instantiation in simulator

        Returns:
            neuron section where the synapse is attached
        """
        if synapse["sectionlist_id"] == 0:
            section = icell.soma[synapse["sectionlist_index"]]
        elif synapse["sectionlist_id"] == 1:
            section = icell.dend[synapse["sectionlist_index"]]
        elif synapse["sectionlist_id"] == 2:
            section = icell.apic[synapse["sectionlist_index"]]
        elif synapse["sectionlist_id"] == 3:
            section = icell.axon[synapse["sectionlist_index"]]

        return section

    def instantiate(self, sim=None, icell=None):
        """Instantiate the synapses.

        In the process, fill the self.pprocesses list.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        if self.rng_settings_mode == "Random123":
            self.rng = sim.neuron.h.Random()
            self.rng.Random123_globalindex(self.seed)

        self.pprocesses = []
        for synapse in self.synapses_data:
            if self.pre_mtypes is None or synapse["pre_mtype"] in self.pre_mtypes:
                # get section
                section = self.get_cell_section_for_synapse(synapse, icell)

                if self.use_glu_synapse:
                    synapse_obj = GluSynapseCustom(
                        sim,
                        icell,
                        synapse,
                        section,
                        self.seed,
                        self.rng_settings_mode,
                        self.synconf_dict,
                    )
                elif self.stim_params is None:
                    synapse_obj = SynapseCustom(
                        sim,
                        icell,
                        synapse,
                        section,
                        self.seed,
                        self.rng_settings_mode,
                        self.synconf_dict,
                    )
                else:
                    stim_params = self.stim_params[synapse["pre_mtype"]]
                    synapse_obj = SynapseCustom(
                        sim,
                        icell,
                        synapse,
                        section,
                        self.seed,
                        self.rng_settings_mode,
                        self.synconf_dict,
                        stim_params[0],  # start
                        stim_params[1],  # interval
                        stim_params[2],  # number
                        stim_params[3],  # noise
                    )

                # setup synapses params for glu synapse case
                if self.use_glu_synapse and self.syn_setup_params is not None:
                    synapse_obj.setup_synapses(self.syn_setup_params)

                self.pprocesses.append(synapse_obj)

    def destroy(self, sim=None):
        """Destroy mechanism instantiation.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unused-argument
        self.pprocesses = None



"""Main Interface."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-error
import tkinter as tk
from tkinter import ttk
import time

from emodelrunner.GUI_utils.simulator import NeuronSimulation
from emodelrunner.GUI_utils.frames import FrameMain, FrameConfig, FrameSynapses
from emodelrunner.GUI_utils.style import define_style, set_matplotlib_style


class GUI:
    """GUI class. Contains all frames and simulation.

    Attributes:
        simulation (NeuronSimulation): contains BluePyOpt simulation (and cell) data
        play (bool): if True, runs the simulation
        refresh_display_dt (float): timestep (s) for the display of figures
        plot_3d (bool): set to True to plot the cell shapes in 3D
        toolbar_on (bool): set to True to display the matplotlib toolbars
        figsize (str): figures size. can be "small", "medium", or "large".
        root (tk.Tk): root of the GUI
        style (ttk.Style): style of the tkinter objects
        frames (dict of ttk.Frames): main frames embedded in root
        reload (bool): if True, the simulation has to be reloaded
    """

    def __init__(self, fps=15, config_path="config/config_allsteps.ini"):
        """Constructor.

        Args:
            fps (int): frames per second for the figure display
            config_path (str): path to the config file used by NeuronSimulation
        """
        # init simulation
        self.simulation = NeuronSimulation(config_path=config_path)

        # load cell, simulation, and protocol(s)
        self.simulation.load_cell_sim()
        self.simulation.load_protocol()

        self.simulation.instantiate()
        self.simulation.load_synapse_display_data()

        # display params
        self.play = False
        self.refresh_display_dt = self.get_refresh_from_fps(fps)
        self.plot_3d = False
        self.toolbar_on = False
        self.figsize = "medium"

        # Tkinter
        self.root = tk.Tk()
        self.root.title("ME-type models launcher & visualisation interface")

        # ttk style
        self.style = ttk.Style()
        self.style.theme_use("clam")  # to be able to change background, etc. on macos
        define_style(self.style)
        set_matplotlib_style()

        # frames
        self.frames = {}
        self.create_frames()

        self.reload = False

    @staticmethod
    def get_refresh_from_fps(fps):
        """Get refresh rate from fps.

        Args:
            fps (float): number of frames per second to be displayed

        Returns:
            float: refresh time (s)
        """
        return 1.0 / fps

    def create_frames(self):
        """Creates the frames."""
        self.frames["FrameMain"] = FrameMain(self.root, self)
        self.frames["FrameConfig"] = FrameConfig(self.root, self)

        title_synapses = ttk.Label(self.root, text="Synapse Stimuli configuration")
        self.frames["FrameSynapses"] = FrameSynapses(self.root, self, title_synapses)

        self.frames["FrameConfig"].grid(
            row=0, column=0, sticky=(tk.W, tk.E, tk.N, tk.S), padx=2, pady=2
        )
        self.frames["FrameMain"].grid(
            row=0, column=1, sticky=(tk.W, tk.E, tk.N, tk.S), pady=2
        )
        self.frames["FrameSynapses"].grid(
            row=0, column=2, sticky=(tk.W, tk.E, tk.N, tk.S), padx=2, pady=2
        )

        self.root.columnconfigure(0, weight=1)
        self.root.columnconfigure(1, weight=1)
        self.root.columnconfigure(2, weight=1)
        self.root.rowconfigure(0, weight=1)

    def update_figures(self):
        """Update the figures."""
        self.frames["FrameMain"].display(self.root, self.simulation)

    def check_v_change(self):
        """Checks the voltage change in the cell sections.

        Update display if change is significant.

        Returns:
            bool: True if the cell morphology with color-coded voltage figure has been updated
        """
        return self.frames["FrameMain"].check_change(self.root, self.simulation)

    def clear_voltage_figure(self):
        """Clear the voltage figure."""
        self.frames["FrameMain"].restart_volt()

    def config_has_changed(self):
        """Stop the simulation when the user has changed configuration."""
        self.reload = True
        self.end_simul()  # stop simul and disable continue button

    def run_simul(self, check_dt=1.5):
        """Main loop for running simulation.

        Args:
            check_dt (float): check for significant voltage change every check_dt (ms)
                (simulation time)
        """
        # for refreshing rate
        t1 = time.time()
        # for fine display of peaks
        last_t = self.simulation.sim.neuron.h.t  # = 0

        while (
            self.play
            and self.simulation.sim.neuron.h.t
            < self.simulation.sim.neuron.h.tstop - self.simulation.sim.neuron.h.dt / 2
        ):
            self.simulation.sim.neuron.h.fadvance()
            if self.simulation.sim.neuron.h.t > last_t + check_dt:
                last_t = self.simulation.sim.neuron.h.t
                # check for big change in voltage. Update display if big change found.
                updated = self.check_v_change()
                if updated:
                    t1 = time.time()
            if time.time() - t1 > self.refresh_display_dt:
                self.update_figures()
                last_t = self.simulation.sim.neuron.h.t
                t1 = time.time()

        self.update_figures()
        self.play = False

        if (
            self.simulation.sim.neuron.h.t
            >= self.simulation.sim.neuron.h.tstop - self.simulation.sim.neuron.h.dt / 2
        ):
            # change buttons state
            self.end_simul()

    def start(self):
        """Start the simulation from beginning. Reload simulation config if needed."""
        # if config has changed: reload cell, sim, protocol, and figure frame
        if self.reload:
            self.reload_params()
        else:
            # destroy last simulation (i.e. this is not a start but a restart)
            if self.simulation.sim.neuron.h.t > 0:
                self.simulation.destroy()
                self.simulation.sim.neuron.h.t = 0
                self.clear_voltage_figure()

            # instantiate
            if self.simulation.cell.icell is None:
                self.simulation.instantiate()

        # change buttons state
        self.frames["FrameMain"].simul_running()

        # run
        self.play = True
        self.run_simul()

    def reload_params(self):
        """Reload cell, protocol, simulation, figure frame."""
        # destroy before reload
        self.simulation.destroy()  # destroy cell, sim and protocol, not config data

        # reload cell & protocol
        self.simulation.load_cell_sim()
        self.simulation.load_protocol()
        self.simulation.instantiate()

        # clear voltage data
        self.clear_voltage_figure()
        # display figures after simulation has been reset
        self.frames["FrameMain"].display(self.root, self.simulation)

        self.root.update()

        # remember that all changes have been loaded
        self.reload = False

    def reload_figure_frame(self):
        """Reload figure frame."""
        self.end_simul()
        self.frames["FrameMain"].destroy()

        # reload figure frame
        self.frames["FrameMain"] = FrameMain(self.root, self)
        self.frames["FrameMain"].grid(
            row=0, column=1, sticky=(tk.W, tk.E, tk.N, tk.S), pady=2
        )
        self.frames["FrameMain"].update_syn_display(self.root, self.simulation)

        self.root.update()

    def pause(self):
        """Pause the simulation."""
        # change buttons state
        self.frames["FrameMain"].simul_on_pause()

        self.play = False

    def continue_simul(self):
        """Unpause the simulation."""
        # change buttons state
        self.frames["FrameMain"].simul_running()

        # run
        self.play = True
        self.run_simul()

    def end_simul(self):
        """End the simulation."""
        self.frames["FrameMain"].simul_ended()
        self.play = False



"""Configuration parsing."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from configparser import ConfigParser
from enum import Enum


class PackageType(Enum):
    """Enumerator for the emodel package types."""

    sscx = "sscx"
    thalamus = "thalamus"
    synplas = "synplas"


class EModelConfigParser(ConfigParser):
    """Built-in ConfigParser annotated with package type."""

    def __init__(self):
        """Constructor."""
        super().__init__()

    @property
    def package_type(self):
        """Package type as a property."""
        return PackageType[self.get("Package", "type")]



"""emodelrunner."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # available since py3.8+
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("emodelrunner")
except PackageNotFoundError:
    # package is not installed
    pass



"""Run pair simulation for synapse plasticity simulation."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging

import numpy as np
from bluepyopt import ephys
from emodelrunner.create_cells import get_precell, get_postcell
from emodelrunner.protocols.create_protocols import define_pairsim_protocols
from emodelrunner.load import get_presyn_stim_args
from emodelrunner.load import get_release_params
from emodelrunner.load import get_syn_setup_params
from emodelrunner.load import load_config
from emodelrunner.output import write_synplas_output
from emodelrunner.output import write_synplas_precell_output
from emodelrunner.run_synplas import _set_global_params

# Configure logger
logger = logging.getLogger(__name__)


def run(
    config_path,
    cvode_active=True,
    postsyn_protocol_name="pulse",
    presyn_protocol_name="presyn_pulse",
    fixhp=True,
):
    """Run cell with pulse stimuli and pre-cell spike train.

    Args:
        config_path (str): path to config file
        cvode_active (bool): whether to use variable time step
        postsyn_protocol_name (str): name of the postsynaptic protocol
        presyn_protocol_name (str): name of the presynaptic protocol
        fixhp (bool): to uninsert SK_E2 for hyperpolarization in cell model
    """
    # pylint:disable=too-many-locals
    config = load_config(config_path=config_path)

    # load extra_params
    syn_setup_params = get_syn_setup_params(
        "synapses/syn_extra_params.json",
        "synapses/cpre_cpost.json",
        config.get("Paths", "synplas_fit_params_path"),
        config.getint("Cell", "gid"),
        config.getboolean("SynapsePlasticity", "invivo"),
    )

    # load cell
    postcell = get_postcell(
        config,
        fixhp=fixhp,
        syn_setup_params=syn_setup_params,
    )

    # load cell
    precell = get_precell(
        config,
        fixhp=fixhp,
    )

    sim = ephys.simulators.NrnSimulator(cvode_active=cvode_active)
    pre_release_params = get_release_params(config, precell=True)
    post_release_params = get_release_params(config)

    # set dynamic timestep tolerance
    sim.neuron.h.cvode.atolscale("v", 0.1)  # 0.01 for more precision

    # load spike_train
    spike_train_path = config.get("Paths", "spiketrain_path")
    pre_spike_train = np.unique(np.loadtxt(spike_train_path, skiprows=1)[:, 0])

    # get pre-synaptic stimulus parameters
    presyn_stim_args = get_presyn_stim_args(config, pre_spike_train)

    # Set fitted model parameters
    if syn_setup_params["fit_params"]:
        _set_global_params(syn_setup_params["fit_params"], sim)

    # Enable in vivo mode (global)
    if config.getboolean("SynapsePlasticity", "invivo"):
        sim.neuron.h.cao_CR_GluSynapse = 1.2  # mM

    # protocols
    protocol = define_pairsim_protocols(
        postcell,
        presyn_protocol_name,
        postsyn_protocol_name,
        cvode_active,
        json.loads(config.get("SynapsePlasticity", "synrec")),
        config.getfloat("Protocol", "tstop"),
        config.getfloat("SynapsePlasticity", "fastforward"),
        presyn_stim_args,
        config.get("Paths", "stimuli_path"),
    )

    # run
    logger.info("Python Recordings Running...")

    responses = protocol.run(
        precell_model=precell,
        postcell_model=postcell,
        pre_param_values=pre_release_params,
        post_param_values=post_release_params,
        sim=sim,
        isolate=False,
    )

    # write responses
    output_path = config.get("Paths", "pairsim_output_path")
    precell_output_path = config.get("Paths", "pairsim_precell_output_path")
    syn_prop_path = config.get("Paths", "syn_prop_path")
    write_synplas_output(responses[1], pre_spike_train, output_path, syn_prop_path)
    write_synplas_precell_output(
        responses[0], presyn_protocol_name, precell_output_path
    )

    logger.info("Python Recordings Done.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config_path",
        default=None,
        help="the path to the config file.",
    )
    args = parser.parse_args()

    run(config_path=args.config_path)



"""Custom Recording class."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

logger = logging.getLogger(__name__)


class SynapseRecordingCustom(ephys.recordings.Recording):
    """Recording in synaptic locations.

    Attributes:
        name (str): name of this object
        location (Location): location in the model of the recording
        variable (str): which variable to record from (e.g. 'v')
        varvectors (list of neuron Vector): vectors recording the variable
        tvector (neuron Vector): vector recording the time (ms)
        instantiated (bool): whether the object has been instantiated or not
    """

    def __init__(self, name=None, location=None, variable="v"):
        """Constructor.

        Args:
            name (str): name of this object
            location (Location): location in the model of the recording
            variable (str): which variable to record from (e.g. 'v')
        """
        super(SynapseRecordingCustom, self).__init__(name=name)
        self.location = location
        self.variable = variable

        self.varvectors = []
        self.tvector = None

        self.instantiated = False

    @property
    def response(self):
        """Return recording responses.

        Returns:
            list of recorded responses
        """
        if not self.instantiated:
            return None

        responses = []
        for varvector in self.varvectors:
            responses.append(
                ephys.recordings.responses.TimeVoltageResponse(
                    self.name, self.tvector.to_python(), varvector.to_python()
                )
            )

        return responses

    def instantiate(self, sim=None, icell=None):
        """Instantiate recording.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        logger.debug(
            "Adding compartment recording of %s at %s", self.variable, self.location
        )

        pprocesses = self.location.instantiate(sim=sim, icell=icell)
        for synapse in pprocesses:
            varvector = sim.neuron.h.Vector()
            varvector.record(getattr(synapse.hsynapse, f"_ref_{self.variable}"))
            self.varvectors.append(varvector)

        self.tvector = sim.neuron.h.Vector()
        self.tvector.record(sim.neuron.h._ref_t)  # pylint: disable=W0212

        self.instantiated = True

    def destroy(self, sim=None):
        """Destroy recording.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        # pylint: disable=unused-argument
        self.varvectors = None
        self.tvector = None
        self.instantiated = False

    def __str__(self):
        """String representation.

        Returns:
            string representation
        """
        return f"{self.name}: {self.variable} at {self.location}"



"""Style-related functions."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib as mpl


def get_style_cst():
    """Returns dict containing style vars such as colors.

    Returns:
        dict: style colors, font and width
    """
    style_dict = {}
    # font & width. has to be an attribute to be accessible.
    # somehow, entry font & width cannot be configurated with style.
    style_dict["base_font"] = "Helvetica 10"
    style_dict["entry_width"] = 8

    # BBP colors
    style_dict["light_blue"] = "#15D3FF"
    style_dict["blue"] = "#0B83CD"
    style_dict["deep_blue"] = "#050A58"
    style_dict["light_grey"] = "#F2F2F2"
    style_dict["grey"] = "#888888"
    style_dict["deep_grey"] = "#333333"
    style_dict["white"] = "#FFFFFF"

    return style_dict


def set_matplotlib_style():
    """Configure ticks & labels size."""
    mpl.rcParams["lines.color"] = get_style_cst()["blue"]
    mpl.rcParams["axes.labelsize"] = 8
    mpl.rcParams["xtick.labelsize"] = 8
    mpl.rcParams["ytick.labelsize"] = 8


def define_style(style):
    """Define the style for ttk objects.

    Args:
        style (ttk.Style): style
    """
    style_dict = get_style_cst()

    style.configure(
        "TButton",
        background=style_dict["white"],
        font=style_dict["base_font"],
    )

    style.configure(
        "ControlSimul.TButton",
        padding=6,
        relief="solid",
        background=style_dict["white"],
        foreground=style_dict["deep_blue"],
        font="Helvetica 16 bold",
        borderwidth=2,
        highlightbackground=style_dict["deep_blue"],  # border color?
    )

    style.map(
        "ControlSimul.TButton",
        foreground=[
            ("pressed", "!disabled", style_dict["blue"]),
            ("disabled", style_dict["grey"]),
        ],
    )

    style.configure("TFrame", background=style_dict["white"])
    style.configure(
        "Boxed.TFrame",
        background=style_dict["white"],
        relief="solid",
        bordercolor=style_dict["deep_blue"],
        borderwidth=4,
    )

    style.configure(
        "TRadiobutton",
        background=style_dict["white"],
        relief="flat",
        cursor="dot",
        borderwidth=0,
        selectcolor=style_dict["blue"],
        font=style_dict["base_font"],
    )

    style.map(
        "TRadiobutton",
        foreground=[
            ("selected", style_dict["blue"]),
            ("!selected", style_dict["deep_blue"]),
        ],
    )

    style.configure(
        "TLabel",
        foreground=style_dict["deep_blue"],
        background=style_dict["white"],
        font=style_dict["base_font"],
    )

    style.configure(
        "TEntry",
        foreground="black",
        background=style_dict["white"],
    )

    style.map(
        "TEntry",
        highlightcolor=[("focus", style_dict["blue"])],
        bordercolor=[("focus", style_dict["blue"])],
    )

    style.configure(
        "TCheckbutton",
        foreground=style_dict["deep_blue"],
        background=style_dict["white"],
        font=style_dict["base_font"],
    )

    style.map(
        "TCombobox",
        fieldbackground=[("!disabled", style_dict["white"])],
        background=[("!disabled", style_dict["white"])],
    )



"""Functions to create synapse locations."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bluepyopt import ephys


def get_syn_locs(cell):
    """Load synapse point process location.

    Args:
        cell (bluepyopt.ephys.models.CellModel): cell model containing the synapses

    Returns:
        list containing the synapse locations
    """
    syn_locs = []
    for mech in cell.mechanisms:
        if hasattr(mech, "pprocesses"):
            syn_locs.append(
                ephys.locations.NrnPointProcessLocation("synapse_locs", mech)
            )

    if not syn_locs:
        syn_locs = None

    return syn_locs



"""Functionality to save factsheets output."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
import numpy as np

from emodelrunner.json_utilities import NpEncoder
from emodelrunner.factsheets.morphology_features import SSCXMorphologyFactsheetBuilder
from emodelrunner.factsheets.physiology_features import physiology_factsheet_info
from emodelrunner.factsheets.experimental_features import get_exp_features_data
from emodelrunner.factsheets.ion_channel_mechanisms import get_mechanisms_data


def write_metype_json(
    data_path,
    current_amplitude,
    stim_start,
    stim_duration,
    morphology_path,
    output_path,
):
    """Write the me-type factsheet json file of SSCX packages.

    The output metype factsheet contains anatomy, physiology and morphology data.

    Args:
        data_path (str): path to the trace data (usually output of emodelrunner run)
        current_amplitude (float): current amplitude of the stimulus (nA)
        stim_start (float): time at which the stimulus begins (ms)
        stim_duration (float): stimulus duration (ms)
        morphology_path (str or Path): Path to the morphology file.
        output_path (str): path to the metype factsheet output
    """
    morphology_path = Path(morphology_path)
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    # load time, voltage
    data = np.loadtxt(data_path)

    morph_factsheet_builder = SSCXMorphologyFactsheetBuilder(morph_path=morphology_path)
    anatomy = morph_factsheet_builder.factsheet_dict()

    physiology = physiology_factsheet_info(
        time=data[:, 0],
        voltage=data[:, 1],
        current_amplitude=current_amplitude,
        stim_start=stim_start,
        stim_duration=stim_duration,
    )
    morphology = {"name": "Morphology name", "value": morphology_path.stem}

    output = [anatomy, physiology, morphology]

    with open(output_path, "w", encoding="utf-8") as out_file:
        json.dump(output, out_file, indent=4, cls=NpEncoder)
    print("me-type json file written.")


def write_etype_factsheet(
    data_path, current_amplitude, stim_start, stim_duration, output_path
):
    """Write the e-type factsheet json file.

    Args:
        data_path (str): path to the trace data (usually output of emodelrunner run)
        current_amplitude (float): current amplitude of the stimulus (nA)
        stim_start (float): time at which the stimulus begins (ms)
        stim_duration (float): stimulus duration (ms)
        output_path (str): path to the etype factsheet output
    """
    data = np.loadtxt(data_path)

    physiology = physiology_factsheet_info(
        time=data[:, 0],
        voltage=data[:, 1],
        current_amplitude=current_amplitude,
        stim_start=stim_start,
        stim_duration=stim_duration,
    )

    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_path, "w", encoding="utf-8") as out_file:
        json.dump(physiology, out_file, indent=4, cls=NpEncoder)


def get_stim_params_from_config_for_physiology_factsheet(prot_path, protocol_key):
    """Get step amplitude, delay and duration for phisiology factsheet.

    Args:
        prot_path (str or Path): path to the json file containing the protocols
        protocol_key (str): name of the protocol used for physiology features extraction

    Returns:
        a tuple containing

        - current_amplitude (int or float): the amplitude current of the step protocol (mA)
        - stim_start (int or float): the start of the stimulus (ms)
        - stim_duration (int or float): the duration of the stimulus (ms)

    Raises:
        Exception: If a step protocol with multiple steps has been provided
    """
    with open(prot_path, "r", encoding="utf-8") as protocol_file:
        protocol_definitions = json.load(protocol_file)

    prot = protocol_definitions[protocol_key]
    step_stim = prot["stimuli"]["step"]

    if isinstance(step_stim, list):
        exception_message = (
            "ME-type factsheet expects only one step stimulus "
            + "for protocol {key} at {filepath}"
        )
        raise Exception(exception_message.format(key=protocol_key, filepath=prot_path))

    # get parameters from protocol
    current_amplitude = step_stim["amp"]
    stim_start = step_stim["delay"]
    stim_duration = step_stim["duration"]

    return current_amplitude, stim_start, stim_duration


def write_metype_json_from_config(
    config, voltage_path, morphology_path, output_path, protocol_key
):
    """Write the me-type factsheet json file from config input.

    Args:
        config (configparser.ConfigParser): configuration
        voltage_path (str): path to the trace data (usually output of emodelrunner run)
        morphology_path (str): Path to the morphology file.
        output_path (str): path to the metype factsheet output
        protocol_key (str): name of the protocol used for physiology features extraction
    """
    # get protocol data
    prot_path = config.get("Paths", "prot_path")

    stim_params = get_stim_params_from_config_for_physiology_factsheet(
        prot_path, protocol_key
    )
    current_amplitude, stim_start, stim_duration = stim_params

    write_metype_json(
        voltage_path,
        current_amplitude,
        stim_start,
        stim_duration,
        morphology_path,
        output_path,
    )


def write_emodel_json(
    emodel,
    morphology_prefix,
    features_dict,
    feature_units_dict,
    unoptimized_params_dict,
    optimized_params_dict,
    output_path,
):
    """Write the e-model factsheet json file.

    The output metype factsheet contains experimental features and channel mechanisms data.

    Args:
        emodel (str): name of the emodel
        morphology_prefix (str): prefix used in the fitness key to the experimental feature
        features_dict (dict): contains the experimental features
        feature_units_dict (dict): contains the units for the experimental features
        unoptimized_params_dict (dict): contains the unoptimized parameters,
            and also contains the decay and exponential equations
        optimized_params_dict (dict): contains the optimized parameters,
            as well as the original morphology path
        output_path (str): path to the e-model factsheet output
    """
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    exp_features = get_exp_features_data(
        emodel,
        morphology_prefix,
        features_dict,
        feature_units_dict,
        optimized_params_dict,
    )
    channel_mechanisms = get_mechanisms_data(
        emodel, optimized_params_dict, unoptimized_params_dict
    )

    output = [
        exp_features,
        channel_mechanisms,
    ]

    with open(output_path, "w", encoding="utf-8") as out_file:
        json.dump(output, out_file, indent=4, cls=NpEncoder)
    print("e-model json file is written.")



"""Physiology features extraction and representation."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import efel


def extract_physiology_features(
    time, voltage, current_amplitude, stim_start, stim_duration
):
    """Extract voltage_base, input_resistance and decay_time_constant features.

    Args:
        time (list): time corresponding to the voltage data of the trace (ms)
        voltage (list): voltage data of the trace (mV)
        current_amplitude (float): current amplitude of the stimulus (nA)
        stim_start (float): time at which the stimulus begins (ms)
        stim_duration (float): stimulus duration (ms)

    Returns:
        list containing voltage base, input resistance and decay time constant features
    """
    # Prepare the trace data
    trace = {}
    trace["T"] = time
    trace["V"] = voltage
    trace["stim_start"] = [stim_start]
    trace["stim_end"] = [stim_start + stim_duration]

    # Calculate the necessary eFeatures
    efel_results = efel.getFeatureValues(
        [trace],
        [
            "voltage_base",
            "steady_state_voltage_stimend",
            "decay_time_constant_after_stim",
        ],
    )

    voltage_base = efel_results[0]["voltage_base"][0]
    dct = efel_results[0]["decay_time_constant_after_stim"][0]

    trace["decay_start_after_stim"] = efel_results[0]["voltage_base"]
    trace["stimulus_current"] = [current_amplitude]

    efel_results = efel.getFeatureValues([trace], ["ohmic_input_resistance_vb_ssse"])
    input_resistance = efel_results[0]["ohmic_input_resistance_vb_ssse"][0]

    return [voltage_base, input_resistance, dct]


def physiology_features_wrapper(voltage_base, input_resistance, dct):
    """Wraps the extracted features into the dictionary format with names and units.

    Args:
        voltage_base (float): resting membrane potential (mV)
        input_resistance (float): input resistance (MOhm)
        dct (float): membrane time constant (ms)

    Returns:
        list containing dicts with each input feature name, value and unit
    """
    factsheet_info = []
    factsheet_info.append(
        {"name": "resting membrane potential", "value": voltage_base, "unit": "mV"}
    )
    factsheet_info.append(
        {"name": "input resistance", "value": input_resistance, "unit": "MOhm"}
    )
    factsheet_info.append(
        {"name": "membrane time constant", "value": dct, "unit": "ms"}
    )
    return factsheet_info


def physiology_factsheet_info(
    time, voltage, current_amplitude, stim_start, stim_duration
):
    """Provides complete physiology information for the factsheet.

    Args:
        time (list): time corresponding to the voltage data of the trace (ms)
        voltage (list): voltage data of the trace (mV)
        current_amplitude (float): current amplitude of the stimulus (nA)
        stim_start (float): time at which the stimulus begins (ms)
        stim_duration (float): stimulus duration (ms)

    Returns:
        dict containing the physiology data
    """
    voltage_base, input_resistance, dct = extract_physiology_features(
        time, voltage, current_amplitude, stim_start, stim_duration
    )
    factsheet_info = physiology_features_wrapper(voltage_base, input_resistance, dct)
    return {"name": "Physiology", "values": factsheet_info}



"""Stimulus class."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bluepyopt.ephys.stimuli import Stimulus


# inspired by bluepyemodel.ecode.sAHP
class Pulse(Stimulus):
    """Train pulse.

    Attributes:
        delay (float): delay after which the stimuli begin (ms)
        duration (float): duration of the stimuli (ms)
        amp (float): amplitude of the stimuli (nA)
        tpulse (float): time between two pulse starts (ms)
        width (float): width of the step stimuli (ms)
        total_duration (float): total duration (delay + duration) (ms)
        location (Location): location of stimulus
        iclamp (neuron IClamp): clamp to inject the stimulus into the cell
        current_vec (neuron Vector): current to inject to the cell
        time_vec (neuron Vector): times at which to play the current
    """

    def __init__(self, location, delay, duration, amp, frequency, width):
        """Constructor.

        Args:
            location (Location): location of stimulus
            delay (float): delay after which the stimuli begin (ms)
            duration (float): duration of the stimuli (ms)
            amp (float): amplitude of the stimuli (nA)
            frequency (float): frequency of the pulse stimuli (1/s)
            width (float): width of the step stimuli (ms)
        """
        self.delay = delay
        self.duration = duration
        self.amp = amp
        self.tpulse = 1000.0 / frequency
        self.width = width
        self.total_duration = self.delay + self.duration

        self.location = location

        self.iclamp = None
        self.current_vec = None
        self.time_vec = None

        super(Pulse, self).__init__()

    def instantiate(self, sim=None, icell=None):
        """Instantiate stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        # adaptation from TStim.hoc proc train from neurodamus

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        if self.width > self.tpulse:
            # one long pulse
            self.time_vec.append(self.delay)
            self.current_vec.append(0)

            self.time_vec.append(self.delay)
            self.current_vec.append(self.amp)

            self.time_vec.append(self.total_duration)
            self.current_vec.append(self.amp)

            self.time_vec.append(self.total_duration)
            self.current_vec.append(0)

        else:
            # repeated pulse injections
            remtime = self.duration
            next_pulse = self.delay

            while remtime > 0:
                if self.width < remtime:
                    self.time_vec.append(next_pulse)
                    self.current_vec.append(0)

                    self.time_vec.append(next_pulse)
                    self.current_vec.append(self.amp)

                    self.time_vec.append(next_pulse + self.width)
                    self.current_vec.append(self.amp)

                    self.time_vec.append(next_pulse + self.width)
                    self.current_vec.append(0)
                else:
                    self.time_vec.append(next_pulse)
                    self.current_vec.append(0)

                    self.time_vec.append(next_pulse)
                    self.current_vec.append(self.amp)

                    self.time_vec.append(self.total_duration)
                    self.current_vec.append(self.amp)

                    self.time_vec.append(self.total_duration)
                    self.current_vec.append(0)

                next_pulse += self.tpulse
                remtime -= self.tpulse

        self.time_vec.append(self.total_duration)
        self.current_vec.append(0)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def destroy(self, sim=None):  # pylint:disable=W0613
        """Destroy stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        self.iclamp = None
        self.time_vec = None
        self.current_vec = None


# inspired by bluepyemodel.ecode.sAHP
class MultipleSteps(Stimulus):
    """Multiple steps protocol at custom times.

    Attributes:
        starts (list): times at which each step occurs (ms)
        amp (float): amplitude of the stimuli (nA)
        width (float): width of the step stimuli (ms)
        total_duration (float): total duration (ms)
        location (Location): location of stimulus
        iclamp (neuron IClamp): clamp to inject the stimulus into the cell
        current_vec (neuron Vector): current to inject to the cell
        time_vec (neuron Vector): times at which to play the current
    """

    def __init__(self, location, starts, amp, width):
        """Constructor.

        Args:
            location (Location): location of stimulus
            starts (list): times at which each step occurs (ms)
            amp (float): amplitude of the stimuli (nA)
            width (float): width of the step stimuli (ms)
        """
        self.starts = starts
        self.amp = amp
        self.width = width
        self.total_duration = self.starts[-1] + self.width

        self.location = location

        self.iclamp = None
        self.current_vec = None
        self.time_vec = None

        super(MultipleSteps, self).__init__()

    def instantiate(self, sim=None, icell=None):
        """Instantiate stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
        """
        # adaptation from TStim.hoc proc train from neurodamus

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        for start in self.starts:
            self.time_vec.append(start)
            self.current_vec.append(0)

            self.time_vec.append(start)
            self.current_vec.append(self.amp)

            self.time_vec.append(start + self.width)
            self.current_vec.append(self.amp)

            self.time_vec.append(start + self.width)
            self.current_vec.append(0)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(0)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def destroy(self, sim=None):  # pylint:disable=W0613
        """Destroy stimulus.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        self.iclamp = None
        self.time_vec = None
        self.current_vec = None



"""Spontaneous Minis."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bluepyopt import ephys


# adapted from bglibpy.cell.add_replay_minis
class Minis(ephys.stimuli.Stimulus):
    """Current stimulus based on stochastic current amplitude."""

    def __init__(
        self,
        gid,
        locations=None,
        stop=None,
        minis_seed=0,
        base_seed=0,
        weight_scalar=None,
        syn_location=None,
        popids=None,
        spont_minis_rate=None,
    ):
        """Constructor."""
        # pylint: disable=too-many-arguments
        super(Minis, self).__init__()
        self.gid = gid
        if stop is None:
            raise ValueError("NrnVecStimStimulus: Need to specify a stop time")
        # must be named total_duration because of ephys.protocols
        self.total_duration = stop

        self.locations = locations

        self.minis_seed = minis_seed
        self.base_seed = base_seed
        self.weight_scalar = weight_scalar
        self.syn_location = syn_location
        self.popids = popids
        self.spont_minis_rate = spont_minis_rate
        self.persistent = []
        self.ips = {}
        self.syn_mini_netcons = {}

    def instantiate(self, sim=None, icell=None):
        """Run stimulus."""
        # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        # pylint: disable=consider-using-f-string
        if self.persistent is None:
            self.persistent = []
        if self.ips is None:
            self.ips = {}
        if self.syn_mini_netcons is None:
            self.syn_mini_netcons = {}

        for location in self.locations:
            # self.connections[location.name] = []
            for synapse in location.instantiate(sim=sim, icell=icell):

                sid = synapse.hsynapse.synapseID

                if self.popids is None:
                    # Default values in Neurodamus
                    source_popid = 0
                    target_popid = 0
                else:
                    source_popid, target_popid = self.popids[sid]

                if self.weight_scalar is not None and sid in self.weight_scalar:
                    weight_scalar = self.weight_scalar[sid]
                else:
                    weight_scalar = 1.0

                if self.spont_minis_rate is not None and sid in self.spont_minis_rate:
                    spont_minis_rate = self.spont_minis_rate[sid]
                else:
                    spont_minis_rate = None

                if spont_minis_rate is not None:
                    # add the *minis*: spontaneous synaptic events
                    self.ips[sid] = sim.neuron.h.InhPoissonStim(
                        self.syn_location[sid], sec=synapse.section  # load this
                    )

                    delay = 0.1
                    self.syn_mini_netcons[sid] = sim.neuron.h.NetCon(
                        self.ips[sid],
                        synapse.hsynapse,
                        -30,
                        delay,
                        synapse.weight * weight_scalar,
                    )
                    # set netcon type
                    nc_param_name = "nc_type_param_{}".format(synapse.hsynapse).split(
                        "[", maxsplit=1
                    )[0]
                    if hasattr(sim.neuron.h, nc_param_name):
                        nc_type_param = int(getattr(sim.neuron.h, nc_param_name))
                        # NC_SPONTMINI
                        self.syn_mini_netcons[sid].weight[nc_type_param] = 1

                    # I assumed rng_settings_mode is the same as in synapse
                    if synapse.rng_settings_mode == "Random123":
                        seed2 = source_popid * 65536 + target_popid + self.minis_seed
                        self.ips[sid].setRNGs(
                            sid + 200,
                            self.gid + 250,
                            seed2 + 300,
                            sid + 200,
                            self.gid + 250,
                            seed2 + 350,
                        )
                    else:
                        exprng = sim.neuron.h.Random()
                        self.persistent.append(exprng)

                        uniformrng = sim.neuron.h.Random()
                        self.persistent.append(uniformrng)

                        if synapse.rng_settings_mode == "Compatibility":
                            exp_seed1 = sid * 100000 + 200
                            exp_seed2 = (
                                self.gid + 250 + self.base_seed + self.minis_seed
                            )
                            uniform_seed1 = sid * 100000 + 300
                            uniform_seed2 = (
                                self.gid + 250 + self.base_seed + self.minis_seed
                            )
                        elif synapse.rng_settings_mode == "UpdatedMCell":
                            exp_seed1 = sid * 1000 + 200
                            exp_seed2 = (
                                source_popid * 16777216
                                + self.gid
                                + 250
                                + self.base_seed
                                + self.minis_seed
                            )
                            uniform_seed1 = sid * 1000 + 300
                            uniform_seed2 = (
                                source_popid * 16777216
                                + self.gid
                                + 250
                                + self.base_seed
                                + self.minis_seed
                            )
                        else:
                            raise ValueError(
                                f"Cell: Unknown rng mode: {synapse.rng_settings_mode}"
                            )

                        exprng.MCellRan4(exp_seed1, exp_seed2)
                        exprng.negexp(1.0)

                        uniformrng.MCellRan4(uniform_seed1, uniform_seed2)
                        uniformrng.uniform(0.0, 1.0)

                        self.ips[sid].setRNGs(exprng, uniformrng)

                    tbins_vec = sim.neuron.h.Vector(1)
                    tbins_vec.x[0] = 0.0
                    rate_vec = sim.neuron.h.Vector(1)
                    rate_vec.x[0] = spont_minis_rate
                    self.persistent.append(tbins_vec)
                    self.persistent.append(rate_vec)
                    self.ips[sid].setTbins(tbins_vec)
                    self.ips[sid].setRate(rate_vec)

    def destroy(self, sim=None):
        """Destroy stimulus."""
        # pylint: disable=unused-argument
        self.persistent = None
        self.ips = None
        self.syn_mini_netcons = None

    def __str__(self):
        """String representation."""
        # pylint: disable=consider-using-f-string
        return (
            "Minis at %s" % ",".join(location for location in self.locations)
            if self.locations is not None
            else "Minis"
        )



# -*- coding: utf-8 -*-
"""Plot shape of neuron."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# pylint: disable=too-many-arguments, too-many-locals, too-many-branches, no-member, import-error
from __future__ import unicode_literals  # for micrometer display
from matplotlib import cm
from neuron.gui2.utilities import _segment_3d_pts


def auto_aspect(ax):
    """Sets the x, y, and z range symmetric around the center.

    Args:
        ax (matplotlib.axes.Axes): axis
    """
    bounds = [ax.get_xlim(), ax.get_ylim()]
    half_delta_max = max([(item[1] - item[0]) / 2 for item in bounds])
    xmid = sum(bounds[0]) / 2
    ymid = sum(bounds[1]) / 2

    ax.set_xlim((xmid - half_delta_max, xmid + half_delta_max))
    ax.set_ylim((ymid - half_delta_max, ymid + half_delta_max))


def get_color_from_cmap(val, val_min, val_max, cmap):
    """Return color.

    Args:
        val (float): value to get color for from colormap (mV)
        val_min (int): minimum value of voltage for colormap (mV)
        val_max (int): minimum value of voltage for colormap (mV)
        cmap (matplotlib.colors.Colormap): colormap

    Returns:
        tuple: tuple of RGBA values indicating a color
    """
    if val_min >= val_max:
        return "black"
    val_range = val_max - val_min
    return cmap((min(max(val, val_min), val_max) - val_min) / (val_range))


def plot_shape(ax, xaxis, yaxis, zaxis, plot_3d, data, linewidth):
    """Plot shape and return line.

    Args:
        ax (matplotlib.axes.Axes): axis
        xaxis (int): indicate which data to plot on the x-axis (0 for x, 1 for y, 2 for z)
        yaxis (int): indicate which data to plot on the y-axis (0 for x, 1 for y, 2 for z)
        zaxis (int): indicate which data to plot on the z-axis (0 for x, 1 for y, 2 for z)
        plot_3d (bool): whether to plot in 3d or 2d
        data (list): list of (xs, ys, zs, diams) for each segment
        linewidth (float): width of line in shape plot

    Returns:
        matplotlib.lines.Line2D: the plotted line
    """
    if plot_3d:
        (line,) = ax.plot(
            data[xaxis],
            data[yaxis],
            data[zaxis],
            "-",
            linewidth=linewidth,
            color="black",
        )
    else:
        (line,) = ax.plot(
            data[xaxis],
            data[yaxis],
            "-",
            linewidth=linewidth,
            color="black",
        )

    return line


def set_labels(ax, xaxis, yaxis, zaxis, plot_3d):
    """Set labels.

    Args:
        ax (matplotlib.axes.Axes): axis
        xaxis (int): indicate which label to set on the x-axis (0 for x, 1 for y, 2 for z)
        yaxis (int): indicate which label to set on the y-axis (0 for x, 1 for y, 2 for z)
        zaxis (int): indicate which label to set on the z-axis (0 for x, 1 for y, 2 for z)
        plot_3d (bool): whether to plot in 3d or 2d
    """
    labels = ["x [μm]", "y [μm]", "z [μm]"]
    ax.set_xlabel(labels[xaxis])
    ax.set_ylabel(labels[yaxis])
    if plot_3d:
        ax.set_zlabel(labels[zaxis])


def get_morph_lines(
    ax,
    sim,
    val_min=-90,
    val_max=30,
    sections=None,
    variable="v",
    cmap=cm.plasma,
    do_plot=False,
    plot_3d=False,
    threshold_volt=4,
    threshold_volt_fine=25,
    old_vals=None,
    vals_last_draw=None,
    xaxis=2,
    yaxis=0,
    zaxis=1,
    linewidth=0.9,
):
    """Plots a 3D shapeplot.

    Args:
        ax(matplotlib.axes.Axes): axis
        sim (bluepyopt.ephys.simulators.NrnSimulator) simulator
        val_min (int): minimum value of voltage for colormap (mV)
        val_max (int): minimum value of voltage for colormap (mV)
        sections (list) list of h.Section() objects to be plotted.
            If None, all sections are loaded.
        variable (str): variable to be plotted. 'v' for voltage.
        cmap (matplotlib.colors.Colormap): colormap
        do_plot (bool): True to plot data. False to get actualised data.
        plot_3d (bool): set to True to plot the shape in 3D
        threshold_volt (int): voltage difference from which
            color should be changed on the cell shape. (mV)
        threshold_volt_fine (int): voltage difference from which
            display should be updated after a small simulation time
            to fine display rapid changes. (mV)
        old_vals (list): variable values at the last display
        vals_last_drawn (list): variable values the last time
            the morphology has been drawn (and not blitted).
        xaxis (int): indicate which data to plot on the x-axis (0 for x, 1 for y, 2 for z)
        yaxis (int): indicate which data to plot on the y-axis (0 for x, 1 for y, 2 for z)
        zaxis (int): indicate which data to plot on the z-axis (0 for x, 1 for y, 2 for z)
        linewidth (float): width of line in shape plot

    Returns:
        tuple containing

        - list: line objects making up shapeplot to update in figure
        - list: voltages previously plotted (mV)
        - bool: True to force the draw of figure, False to blit the figure
    """
    # Adapted from the NEURON package (fct _do_plot in __init__):
    # https://www.neuron.yale.edu/neuron/
    # where this part of the code was itself adapted from
    # https://github.com/ahwillia/PyNeuron-Toolbox/blob/master/PyNeuronToolbox/morphology.py
    # Accessed 2019-04-11, which had an MIT license

    # Default is to plot all sections.
    if sections is None:
        sections = list(sim.neuron.h.allsec())

    sim.neuron.h.define_shape()

    if do_plot:
        lines_list = []
    else:
        # get lines to be actualised
        lines_list = ax.lines

    lines_to_update = []
    vals = []

    # get lines and variable values at each segment
    for i, sec in enumerate(sections):
        # Plot each segment as a line
        if do_plot:
            all_seg_pts = _segment_3d_pts(sec)

            for seg, data in zip(sec, all_seg_pts):
                val = getattr(seg, variable)
                vals.append(val)

                line = plot_shape(ax, xaxis, yaxis, zaxis, plot_3d, data, linewidth)
                set_labels(ax, xaxis, yaxis, zaxis, plot_3d)

                if cmap:
                    col = get_color_from_cmap(val, val_min, val_max, cmap)
                    line.set_color(col)
                lines_list.append(line)

        else:
            for seg in sec:
                val = getattr(seg, variable)
                vals.append(val)

    if do_plot:
        auto_aspect(ax)

    if old_vals is None:
        old_vals = [100] * len(vals)

    force_draw = False
    if old_vals and cmap:
        for i, (val, old_val) in enumerate(zip(vals, old_vals)):
            if val is not None and abs(val - old_val) > threshold_volt:
                col = get_color_from_cmap(val, val_min, val_max, cmap)

                lines_list[i].set_color(col)
                lines_to_update.append(lines_list[i])

                old_vals[i] = val

                if (
                    vals_last_draw is not None
                    and abs(val - vals_last_draw[i]) > threshold_volt_fine
                ):
                    force_draw = True

    if force_draw:
        lines_to_update = lines_list

    return lines_to_update, old_vals, force_draw



"""Functions to write output."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json

import h5py
import numpy as np


def write_responses(responses, output_dir):
    """Write each response in a file.

    Args:
        responses (dict): time and recorded value of each recording
            Should have structure "key": {"time": time, "voltage": response}
            Note that all response have a "voltage" field, even if the recorded value
            was not the voltage
        output_dir (str): path to the output repository
    """
    for key, resp in responses.items():
        output_path = os.path.join(output_dir, key + ".dat")

        # Some resp are None when spike is not found
        if resp is None:
            continue
        if isinstance(resp, (float, np.float)):
            np.savetxt(output_path, np.array([resp]))
        else:
            time = np.array(resp["time"])
            soma_voltage = np.array(resp["voltage"])

            np.savetxt(output_path, np.transpose(np.vstack((time, soma_voltage))))


def write_current(currents, output_dir):
    """Write currents into separate files.

    Args:
        currents (dict): time and trace to each recording
            Should have structure "key": {"time": time, "current": current}
        output_dir (str): path to the output repository
    """
    for key, curr_dict in currents.items():
        output_path = os.path.join(output_dir, key + ".dat")
        np.savetxt(
            output_path,
            np.transpose(np.vstack((curr_dict["time"], curr_dict["current"]))),
        )


def write_synplas_output(
    responses,
    pre_spike_train,
    output_path="./output.h5",
    syn_prop_path="synapses/synapse_properties.json",
):
    """Write output as h5.

    Args:
        responses (dict): responses of the postsynaptic cell
        pre_spike_train (list): times at which the synapses fire (ms)
        output_path (str): path to the (postsynaptic data) output file
        syn_prop_path (str): path to the synapse properties file
    """
    results = {"prespikes": pre_spike_train}
    # add synprop
    if os.path.isfile(syn_prop_path):
        with open(syn_prop_path, "r", encoding="utf-8") as f:
            synprop = json.load(f)
            results["synprop"] = synprop

    # add responses
    for key, resp in responses.items():
        if isinstance(resp, list):
            results[key] = np.transpose([np.array(rec["voltage"]) for rec in resp])
        else:
            results["t"] = np.array(resp["time"])
            results["v"] = np.array(resp["voltage"])

    # Store results
    h5file = h5py.File(output_path, "w")
    for key, result in results.items():
        if key == "synprop":
            h5file.attrs.update(results["synprop"])
        else:
            h5file.create_dataset(
                key,
                data=result,
                chunks=True,
                compression="gzip",
                compression_opts=9,
            )
    h5file.close()


def write_synplas_precell_output(
    responses,
    protocol_name,
    precell_output_path="./output_precell.h5",
):
    """Write precell output as h5.

    Args:
        responses (dict): responses of the presynaptic cell
        protocol_name (str): name of the presynaptic protocol
        precell_output_path (str): path to the presynaptic data output file
    """
    results = {}

    # add responses
    results["t"] = np.array(responses[protocol_name]["time"])
    results["v"] = np.array(responses[protocol_name]["voltage"])

    # Store results
    h5file = h5py.File(precell_output_path, "w")
    for key, result in results.items():
        h5file.create_dataset(
            key,
            data=result,
            chunks=True,
            compression="gzip",
            compression_opts=9,
        )
    h5file.close()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))

try:
    # available since py3.8+
    from importlib.metadata import version as get_version
except ModuleNotFoundError:
    from importlib_metadata import version as get_version

from pathlib import Path
import shutil

# -- copy images referenced in Readme ---------------------------------------

Path("doc/source/images").mkdir(parents=True, exist_ok=True)
shutil.copy("images/GUI_screenshot.png", "doc/source/images/GUI_screenshot.png")


# -- Project information -----------------------------------------------------

project = "EModelRunner"

release = get_version("emodelrunner")
version = release


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autosummary_generate = True
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    # references twice, in protocols.RatSSCxMainProtocol
    # once in Attributes, and once in a property method
    "exclude-members": "rin_efeature",
}
# so that we don't have to install neuron matplotlib, tkinter
# (imported in GUI) to do the docs
autodoc_mock_imports = ["neuron", "matplotlib", "tkinter"]
# to be able to put multiple return variables in the docstrings
# napoleon_custom_sections = [("Returns", "params_style")]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    "metadata_distribution": "emodelrunner",
}

html_title = "EModelRunner"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False



"""Custom synapse-related classes."""

# Copyright 2020-2021 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SynapseMixin:
    """Class containing the synapse-related methods."""

    def set_random_nmb_generator(self, sim, icell, sid):
        """Sets the random number generator.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
            sid (int): synapse id
        """
        if self.rng_settings_mode == "Random123":
            self.randseed1 = icell.gid + 250
            self.randseed2 = sid + 100
            self.randseed3 = 300

            self.hsynapse.setRNG(self.randseed1, self.randseed2, self.randseed3)

        if self.rng_settings_mode == "Compatibility":
            self.rndd = sim.neuron.h.Random()
            self.rndd.MCellRan4(
                sid * 100000 + 100,
                icell.gid + 250 + self.seed,
            )
            self.rndd.uniform(0, 1)

            self.hsynapse.setRNG(self.rndd)

    def set_tau_r(self, sim, icell, sid):
        """Set tau_r_GABAA using random nmb generator.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            icell (neuron cell): cell instantiation in simulator
            sid (int): synapse id
        """
        self.rng = sim.neuron.h.Random()
        if self.rng_settings_mode == "Random123":
            self.rng.Random123(icell.gid + 250, sid + 100, 450)
        elif self.rng_settings_mode == "Compatibility":
            self.rng.MCellRan4(
                sid * 100000 + 100,
                icell.gid + 250 + self.seed,
            )
        self.rng.lognormal(0.2, 0.1)
        self.hsynapse.tau_r_GABAA = self.rng.repick()

    def execute_synapse_configuration(self, synconf_dict, sid, sim, exec_all=False):
        """Create a hoc file configuring synapse.

        Args:
            synconf_dict (dict): synapse configuration
            sid (int): synapse id
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            exec_all (bool): whether to also execute commands with '*'
        """
        # pylint: disable=consider-using-f-string
        for cmd, ids in synconf_dict.items():
            if sid in ids and (exec_all or "*" not in cmd):
                cmd = cmd.replace("%s", "\n%(syn)s")
                hoc_cmd = cmd % {"syn": self.hsynapse.hname()}
                hoc_cmd = "{%s}" % hoc_cmd
                sim.neuron.h(hoc_cmd)


class SynapseCustom(SynapseMixin):
    """Attach a synapse to the simulation.

    Attributes:
        seed (int): random number generator seed number
        rng_settins_mode (str) : mode of the random number generator
            Can be "Random123" or "Compatibility"
        section (neuron section): cell location where the synapse is attached to
        hsynapse (neuron ProbGABAAB_EMS or ProbAMPANMDA_EMS): synapse instantion in simulator
        delay (float): synapse delay
        weight (float): synapse weight
        pre_mtype (int): ID (but not gid) of the presynaptic cell
        start (int/None): force synapse to start firing at given value when using NetStim
        interval (int/None): force synapse to fire at given interval when using NetStim
        number (int/None): force synapse to fire N times when using NetStim
        noise (int/None): force synapse to have given noise when using NetStim
    """

    def __init__(
        self,
        sim,
        icell,
        synapse,
        section,
        seed,
        rng_settings_mode,
        synconf_dict,
        start=None,
        interval=None,
        number=None,
        noise=None,
    ):
        """Constructor.

        Args:
            sim (NrnSimulator): simulator
            icell (Hoc Cell): cell to which attach the synapse
            synapse (dict): synapse data
            section (neuron section): cell location where the synapse is attached to
            seed (int) : random number generator seed number
            rng_settings_mode (str) : mode of the random number generator
                Can be "Random123" or "Compatibility"
            synconf_dict (dict) : synapse configuration
            start (int/None): force synapse to start firing at given value when using NetStim
            interval (int/None): force synapse to fire at given interval when using NetStim
            number (int/None): force synapse to fire N times when using NetStim
            noise (int/None): force synapse to have given noise when using NetStim
        """
        # pylint: disable=too-many-arguments
        self.seed = seed
        self.rng_settings_mode = rng_settings_mode
        self.section = section

        # the synapse is inhibitory
        if synapse["synapse_type"] < 100:
            self.hsynapse = sim.neuron.h.ProbGABAAB_EMS(
                synapse["seg_x"], sec=self.section
            )
            self.hsynapse.tau_d_GABAA = synapse["tau_d"]

            self.set_tau_r(sim, icell, synapse["sid"])
        # the synapse is excitatory
        else:
            self.hsynapse = sim.neuron.h.ProbAMPANMDA_EMS(
                synapse["seg_x"], sec=self.section
            )
            self.hsynapse.tau_d_AMPA = synapse["tau_d"]

        self.hsynapse.Use = abs(synapse["use"])
        self.hsynapse.Dep = abs(synapse["dep"])
        self.hsynapse.Fac = abs(synapse["fac"])

        self.hsynapse.synapseID = synapse["sid"]

        self.hsynapse.Nrrp = synapse["Nrrp"]

        # set random number generator
        self.set_random_nmb_generator(sim, icell, synapse["sid"])

        self.execute_synapse_configuration(synconf_dict, synapse["sid"], sim)

        self.delay = synapse["delay"]
        self.weight = synapse["weight"]

        self.pre_mtype = synapse["pre_mtype"]

        # netstim params if given
        self.start = start
        self.interval = interval
        self.number = number
        self.noise = noise
